import json
import os


def open_fav_folder_with(
    bookmarks_path: str,
    root_folder: str,
    fav_folder: str,
    start_proto: str = None,
):
    bookmarks = None
    with open(bookmarks_path, encoding="utf8") as f:
        bookmarks = json.load(f)

    # open folder recursively
    def open_folder(folder):
        for child in folder["children"]:
            if child.get("type") == "url":
                url = child["url"]
                if start_proto:
                    url = f"{start_proto}:{url}"
                os.startfile(url)
            else:
                open_folder(child)

    # locate the initial folder
    folder = bookmarks["roots"][root_folder]
    for component in fav_folder.split("."):
        for child in folder["children"]:
            if child["name"] == component and child["type"] == "folder":
                folder = child
                break
    open_folder(folder)


def open_firefox_fav_folder(places_path, fav_folder='daily'):
    import sqlite3
    sql_query = f"""
        select
            -- mb.title,
            mp.url
        from
            moz_bookmarks mb
        left join moz_places mp on
            (mb.fk = mp.id)
        where
            mb.parent =(
            select
                id
            from
                moz_bookmarks
            where
                title = ?
                and type = 2)   
    """
    con = sqlite3.connect(places_path)
    cur = con.cursor()
    res = cur.execute(sql_query, [fav_folder])
    for url, in res.fetchall():
        os.startfile(url)


def open_chrome_fav_folder(root_folder, fav_folder):
    """Opens the Chrome Favorites folder"""
    bookmarks_path = os.path.join(
        os.getenv("LOCALAPPDATA"),
        "Google",
        "Chrome",
        "User Data",
        "Default",
        "Bookmarks",
    )
    return open_fav_folder_with(bookmarks_path, root_folder, fav_folder)


def open_edge_fav_folder(root_folder, fav_folder):
    """Opens the Chrome Favorites folder"""
    bookmarks_path = os.path.join(
        os.getenv("LOCALAPPDATA"),
        "Microsoft",
        "Edge",
        "User Data",
        "Default",
        "Bookmarks",
    )
    return open_fav_folder_with(
        bookmarks_path, root_folder, fav_folder, "microsoft-edge"
    )



import logging
from typing import Dict, List, Optional, Set
from os import path

from jigsawwm.tiler.tilers import *
from jigsawwm.w32.monitor import Monitor, get_monitor_from_window
from jigsawwm.w32.window import RECT, Window, get_active_window

from .theme import Theme

logger = logging.getLogger(__name__)


class MonitorState:
    """MonitorState holds variables needed by a Monitor


    :param VirtDeskState virtdesk_state: associated virtual desktop
    :param Monitor monitor: associated system monitor
    :param str theme: the active theme for the monitor in the virtual desktop
    """

    virtdesk_state: "VirtDeskState"
    monitor: Monitor
    theme: Optional[str]
    windows: List[Window]
    last_active_window: Optional[Window] = None

    def __init__(
        self,
        virtdesk_state: "VirtDeskState",
        monitor: Monitor,
        theme: Optional[str] = None,
    ):
        self.virtdesk_state = virtdesk_state
        self.monitor = monitor
        self.theme = theme
        self.windows = []
        self.last_active_window = None
        self.hook_ids = []

    def get_existing_windows(self) -> List[Window]:
        """Retrieves current managed windows"""
        return [w for w in self.windows if w.exists()]

    def sync(self, windows: Set[Window], restrict=False, window_sort_order=[]):
        """Synchronize managed windows with given actual windows currently visible and arrange them accordingly

        :param Set[Window] windows: latest visible windows
        :param bool restrict: optional, restrict windows to their specified rect no matter what
        """
        theme = self.virtdesk_state.get_theme(self.theme)


        #
        # remove invalid windows from list
        #

        old_list = self.windows
        new_list = []
        for w in old_list:
            if w not in windows:
                continue
            windows.remove(w)
            if not w.exists():
                continue
            new_list.append(w)


        #
        # prepend or append the new windows
        #

        if theme.new_window_as_master:
            new_list = list(windows) + new_list
        else:
            new_list = new_list + list(windows)


        #
        # sort windows according to init sequence
        #

        _i = 0
        sorted_list = []
        if len(window_sort_order) > 0:

            # over all window elements of desired sequence
            for w_seq in window_sort_order:
                # over all currently (to become) active window objects
                for w_cur in new_list:
                    # check for correspondence
                    if w_seq[0].lower() == path.basename(w_cur.exe).lower() \
                            and ( len(w_seq) < 2 or not w_seq[1] or w_seq[1].lower() in w_cur.title.lower() ):
                        # add window object at prioritized position
                        sorted_list.append(w_cur)
                        # remove window object from source list
                        new_list.remove(w_cur)
                        # continue with next window from sequence
                        break

            # compose new list of sorted part and remainder
            new_list = sorted_list + new_list


        #
        # skip if there is nothing changed unless strict mode
        #

        if new_list == old_list:
            if restrict:
                self.restrict(theme)
            return


        self.windows = new_list
        self.arrange(theme)

    def arrange(self, theme: Optional[Theme] = None):
        """Arrange windows based on the theme

        :param str theme: optional, fallback to theme of the instance
        """
        theme = theme or self.virtdesk_state.get_theme(self.theme)
        print("arrange", self.theme, theme.name)
        wr = self.monitor.get_info().rcWork
        work_area = (wr.left, wr.top, wr.right, wr.bottom)
        windows = self.get_existing_windows()
        i = 0
        gap = theme.gap
        for left, top, right, bottom in theme.layout_tiler(work_area, len(windows)):
            window = windows[i]
            # add gap
            if gap:
                if left == wr.left:
                    left += gap
                if top == wr.top:
                    top += gap
                if right == wr.right:
                    right -= gap
                if bottom == wr.bottom:
                    bottom -= gap
            left += gap
            top += gap
            right -= gap
            bottom -= gap
            rect = RECT(left, top, right, bottom)
            logger.debug("arrange %s %s", window, rect)
            window.set_rect(rect)
            # compensation
            r = window.get_rect()
            b = window.get_extended_frame_bounds()
            compensated_rect = (
                round(left + r.left - b.left),
                round(top + r.top - b.top),
                round(right + r.right - b.right),
                round(bottom + r.bottom - b.bottom),
            )
            window.set_rect(RECT(*compensated_rect))
            i += 1

    def restrict(self, theme: Optional[Theme] = None):
        """Restrict all managed windows to their specified rect"""
        theme = theme or self.get_theme()
        if not theme.strict:
            return
        for window in self.windows:
            window.restrict()


class VirtDeskState:
    """VirtDeskState holds variables needed by a Virtual Desktop

    :param WindowManager manager: associated WindowManager
    :param bytearray desktop_id: virtual desktop id
    """

    desktop_id: bytearray
    managed_windows: Set[Window]
    monitors: Dict[Monitor, MonitorState]
    last_active_window: Optional[Window] = None

    def __init__(self, get_theme: Callable[[str], Theme], desktop_id: bytearray):
        self.desktop_id = desktop_id
        self.managed_windows = set()
        self.monitors = {}
        self.last_active_window = None
        self.get_theme = get_theme

    def get_monitor(self, monitor: Monitor) -> MonitorState:
        """Retrieves the monitor state for the specified monitor in the virtual desktop

        :param Monitor monitor: monitor
        :returns: monitor state
        :rtype: MonitorState
        """
        monitor_state = self.monitors.get(monitor)
        if monitor_state is None:
            monitor_state = MonitorState(self, monitor)
            self.monitors[monitor] = monitor_state
        return monitor_state

    def get_managed_active_window(self) -> Optional[Window]:
        """Retrieves the managed forground window if any"""
        window = get_active_window()
        logger.debug("get_managed_active_window: active window %s", window)
        if window is None:
            return None
        if window not in self.managed_windows:
            logger.debug(
                "get_managed_active_window: active window is NOT managed %s", window
            )
            return None
        return window

    def get_last_managed_active_window(self) -> Optional[Window]:
        """Retrieves the latest managed forground window if any"""
        if self.last_active_window and (
            self.last_active_window not in self.managed_windows
            or not self.last_active_window.exists()
        ):
            return None
        return self.last_active_window

    def find_owner(self, window: Window) -> Optional[MonitorState]:
        """Retrieves the windows list containing specified window and its index in the list"""
        monitor = get_monitor_from_window(window.handle)
        monitor_state = self.get_monitor(monitor)
        if window in monitor_state.windows:
            return monitor_state
        return None



import logging
from functools import partial
from os import path
from typing import Dict, List, Optional, Set

from jigsawwm import ui
from jigsawwm.tiler.tilers import *
from jigsawwm.w32 import hook
from jigsawwm.w32.monitor import (
    Monitor,
    get_monitor_from_cursor,
    get_monitor_from_window,
    get_topo_sorted_monitors,
    set_cursor_pos,
)
from jigsawwm.w32.reg import get_current_desktop_id
from jigsawwm.w32.window import (
    DWORD,
    HWND,
    LONG,
    Window,
    get_foreground_window,
    get_manageable_windows,
    get_window_from_pos,
    get_window_title,
    is_app_window,
    is_window,
)
from jigsawwm.w32.winevent import WinEvent

from .op_mixin import OpMixin
from .state import MonitorState, VirtDeskState
from .theme import Theme

logger = logging.getLogger(__name__)


class WindowManager(OpMixin):
    """WindowManager detect the monitors/windows state and arrange them dynamically
    keep in mind that not all windows are managed by the WindowManager, only those
    returned by `jigsawwm.w32.get_normal_windows()` and not ignored would be managed.

    The WindowManager works just like other Dynamic Window Managers, it store all
    managed windows in a list, the first one in the list is called Master, it would
    normally take up most area of the screen while others occuppy the rest.

    :param List[Theme] themes: all avaiable themes for user to switch
    :param ignore_exe_names: list of executable filenames that you don't want them
                             to be managed/arranged
    :param init_exe_sequence: list of executable filenames and title search criteria
                              that are to be kept in exactly this order when
                              distributing into windows
    """

    _state: Dict[bytearray, VirtDeskState]
    themes: List[Theme]
    ignore_exe_names: Set[str]
    force_managed_exe_names: Set[str]
    init_exe_sequence: List[List[str]]

    def __init__(
        self,
        themes: List[Theme] = None,
        ignore_exe_names: Set[str] = None,
        force_managed_exe_names: Set[str] = None,
        init_exe_sequence: List[List[str]] = None,
    ):
        self._state = {}
        self.themes = themes or [
            Theme(
                name="WideScreen Dwindle",
                layout_tiler=widescreen_dwindle_layout_tiler,
                icon_name="wide-dwindle.png",
            ),
            Theme(
                name="OBS Dwindle",
                layout_tiler=obs_dwindle_layout_tiler,
                icon_name="obs.png",
            ),
            Theme(
                name="Dwindle",
                layout_tiler=dwindle_layout_tiler,
                icon_name="dwindle.png",
            ),
        ]
        self.ignore_exe_names = set(ignore_exe_names or [])
        self.force_managed_exe_names = set(force_managed_exe_names or [])
        self.init_exe_sequence = init_exe_sequence or []
        self.theme = self.themes[0].name
        self.sync(init=True)

    @property
    def virtdesk_state(self) -> Optional[VirtDeskState]:
        """Retrieve virtual desktop state"""
        desktop_id = get_current_desktop_id()
        virtdesk_state = self._state.get(desktop_id)
        if virtdesk_state is None:
            # make sure monitor_state for current virtual desktop exists
            virtdesk_state = VirtDeskState(
                lambda theme: self.themes[self.get_theme_index(theme)], desktop_id
            )
            self._state[desktop_id] = virtdesk_state
        return virtdesk_state

    def check_window_ignored(self, window: Window) -> bool:
        exepath = window.exe
        return not exepath or path.basename(exepath) in self.ignore_exe_names

    def check_force_managed(self, hwnd: HWND) -> bool:
        try:
            exepath = Window(hwnd).exe
            return exepath and path.basename(exepath) in self.force_managed_exe_names
        except:
            return False

    def sync(self, init=False, restrict=False) -> bool:
        """Update manager state(monitors, windows) to match OS's and arrange windows if it is changed"""

        virtdesk_state = self.virtdesk_state


        #
        # gather all manageable windows
        #

        manageable_windows = list(get_manageable_windows(self.check_force_managed))
        if not manageable_windows:
            return


        #
        # group manageable windows by their current monitor
        #

        group_wins_by_mons: Dict[Monitor, Set[Window]] = {}
        managed_windows = set()
        for window in manageable_windows:


            #
            # skip this window if to be ignored
            #

            if self.check_window_ignored(window):
                continue


            #
            # determine relevant monitor
            #

            if init or window in virtdesk_state.managed_windows:

                # use previous monitor
                monitor = get_monitor_from_window(window.handle)

            else:
                # use the monitor currently showing the cursor
                monitor = get_monitor_from_cursor()


            #
            # build list of windows for each monitor
            #

            # get current list of windows for relevant monitor
            windows = group_wins_by_mons.get(monitor)

            # init list if not yet existing
            if windows is None:
                windows = set()
                group_wins_by_mons[monitor] = windows

            # add window to lists
            windows.add(window)
            managed_windows.add(window)
            logger.debug("%s is managed by monitor %s", window, monitor)


        #
        # synchronize windows on each monitor
        #

        virtdesk_state.managed_windows = managed_windows

        # pass down to monitor_state for further synchronization
        for monitor, windows in group_wins_by_mons.items():
            monitor_state = virtdesk_state.get_monitor(monitor)
            monitor_state.sync(
                    windows,
                    restrict=restrict,
                    #window_sort_order=self.init_exe_sequence if init else [],
                    window_sort_order=self.init_exe_sequence,
                    )

    def arrange_all_monitors(self):
        """Arrange all windows in all monitors to where their suppose to be"""
        virtdesk_state = self.virtdesk_state
        for monitor in virtdesk_state.monitors.values():
            monitor.arrange()

    def activate(self, window: Window):
        """Activate specified window"""
        virtdesk_state = self.virtdesk_state
        window.activate()
        # move cursor to the center of the window
        rect = window.get_rect()
        set_cursor_pos(
            rect.left + (rect.right - rect.left) / 2,
            rect.top + (rect.bottom - rect.top) / 2,
        )
        virtdesk_state.last_active_window = window
        monitor = get_monitor_from_window(window.handle)
        virtdesk_state.get_monitor(monitor).last_active_window = window

    def activate_by_offset(self, offset: int):
        """Activate managed window by offset

        When the active window is managed, activate window in the same monitor by offset
        When the active window is unmanaged, activate the last active window
        When none of above viable, activate the Master window in the montior under cursor
        Or, do nothing!
        """
        virtdesk_state = self.virtdesk_state
        src_window = virtdesk_state.get_managed_active_window()
        dst_window = None
        if src_window:
            monitor_state = virtdesk_state.find_owner(src_window)
            if len(monitor_state.windows) > 0:
                src_index = monitor_state.windows.index(src_window)
                dst_index = (src_index + offset) % len(monitor_state.windows)
                dst_window = monitor_state.windows[dst_index]
        else:
            dst_window = virtdesk_state.last_active_window
        if not dst_window:
            monitor_state = virtdesk_state.get_monitor(get_monitor_from_cursor())
            if len(monitor_state.windows) > 0:
                dst_window = monitor_state.windows[0]
        if not dst_window:
            return
        self.activate(dst_window)
        ui.show_windows_splash(monitor_state, dst_window)

    def _reorder(self, reorderer: Callable[[List[Window], int], None]):
        virtdesk_state = self.virtdesk_state
        src_window = virtdesk_state.get_managed_active_window()
        if not src_window:
            return
        monitor_state = virtdesk_state.find_owner(src_window)
        if monitor_state is None:
            return
        if len(monitor_state.windows) < 2:
            return
        reorderer(monitor_state.windows, monitor_state.windows.index(src_window))
        monitor_state.arrange()
        self.activate(src_window)

    def swap_by_offset(self, offset: int):
        """Swap current active managed window with its sibling by offset"""

        def reorderer(windows: List[Window], src_idx: int):
            dst_idx = (src_idx + offset) % len(windows)
            windows[src_idx], windows[dst_idx] = windows[dst_idx], windows[src_idx]

        self._reorder(reorderer)

    def swap_master(self):
        """Swap the current active managed window with the Master or the second window
        in the list if it is Master already
        """

        def reorderer(windows: List[Window], src_idx: int):
            if src_idx == 0:
                dst_idx = 1
            else:
                dst_idx = 0
            windows[src_idx], windows[dst_idx] = windows[dst_idx], windows[src_idx]

        self._reorder(reorderer)

    def set_master(self):
        """Set the active active managed window as the Master or the second window
        in the list if it is Master already
        """

        def reorderer(windows: List[Window], src_idx: int):
            src_window = windows[src_idx]
            if src_idx == 0:
                src_idx = 1
                src_window = windows[1]
            # shift elements from the beginning to the src_window
            for i in reversed(range(1, src_idx + 1)):
                windows[i] = windows[i - 1]
            # assign new master
            windows[0] = src_window

        self._reorder(reorderer)

    def get_theme_index(self, theme_name: str) -> int:
        """Retrieves the index of given theme name, useful to switching theme"""
        i = len(self.themes) - 1
        while i > 0:
            if self.themes[i].name == theme_name:
                return i
            i -= 1
        return i

    def switch_theme_by_offset(self, delta: int):
        """Switch theme by offset"""
        try:
            virtdesk_state = self.virtdesk_state
            monitor = (
                get_monitor_from_window(get_foreground_window())
                or get_monitor_from_cursor()
            )
            monitor_state = virtdesk_state.get_monitor(monitor)
            theme_index = self.get_theme_index(monitor_state.theme)
            theme = self.themes[(theme_index + delta) % len(self.themes)]
            # new_theme_name = self.themes[(theme_index + delta) % len(self.themes)].name
            monitor_state.theme = theme.name
            monitor_state.arrange(theme)
        except:
            import traceback

            traceback.print_exc()

    def get_monitor_state_pair(
        self, delta: int, virtdesk_state: Optional[VirtDeskState] = None
    ) -> Tuple[MonitorState, MonitorState]:
        """Retrieves a pair of monitor_states, from cursor and its offset in the list"""
        virtdesk_state = virtdesk_state or self.virtdesk_state
        if not virtdesk_state:
            return
        monitors = get_topo_sorted_monitors()
        src_monitor = get_monitor_from_cursor()
        src_idx = monitors.index(src_monitor)
        dst_idx = (src_idx + delta) % len(monitors)
        dst_monitor = monitors[dst_idx]
        dst_monitor_state = virtdesk_state.get_monitor(dst_monitor)
        src_monitor_state = virtdesk_state.get_monitor(src_monitor)
        return src_monitor_state, dst_monitor_state

    def switch_monitor_by_offset(self, delta: int):
        """Switch to another monitor by given offset"""
        logger.debug("switch_monitor_by_offset: %s", delta)
        _, dst_monitor_state = self.get_monitor_state_pair(delta)
        rect = dst_monitor_state.monitor.get_info().rcWork
        x, y = (
            rect.left + (rect.right - rect.left) / 2,
            rect.top + (rect.bottom - rect.top) / 2,
        )
        set_cursor_pos(x, y)
        window = get_window_from_pos(x, y)
        if not window:
            window = dst_monitor_state.last_active_window
        if window is None or not window.exists():
            windows = dst_monitor_state.get_existing_windows()
            if windows:
                window = windows[0]
        if window:
            self.activate(window)

    def move_to_monitor_by_offset(self, delta: int):
        """Move active window to another monitor by offset"""
        logger.debug("move_to_monitor_by_offset(%s)", delta)
        virtdesk_state = self.virtdesk_state
        window = virtdesk_state.get_managed_active_window()
        if not window:
            return
        src_monitor_state, dst_monitor_state = self.get_monitor_state_pair(
            delta, virtdesk_state
        )
        src_monitor_state.windows.remove(window)
        src_monitor_state.arrange()
        dst_monitor_state.windows.append(window)
        dst_monitor_state.arrange()
        # idx = src_monitor_state.windows.index(window)
        # src_win_len = len(src_monitor_state.windows)
        # if src_win_len:
        #     next_window = src_monitor_state.windows[idx % src_win_len]
        #     self.activate(next_window)
        self.activate(window)

    def _winevent_callback(
        self,
        event: WinEvent,
        hwnd: HWND,
        id_obj: LONG,
        id_chd: LONG,
        id_evt_thread: DWORD,
        evt_time: DWORD,
        restrict: bool = False,
    ):
        if (
            id_obj
            or id_chd
            or not is_window(hwnd)
            or not is_app_window(hwnd)
            or self.check_window_ignored(Window(hwnd))
        ):
            return
        logger.debug(
            "_winevent_callback: event %s restrict %s %s",
            event.name,
            restrict,
            get_window_title(hwnd),
        )
        self.sync(restrict=restrict)

    def install_hooks(self):
        """Install hooks for window events"""
        self.hook_ids = [
            hook.hook_winevent(
                WinEvent.EVENT_OBJECT_SHOW,
                WinEvent.EVENT_OBJECT_HIDE,
                self._winevent_callback,
            ),
            hook.hook_winevent(
                WinEvent.EVENT_OBJECT_CLOAKED,
                WinEvent.EVENT_OBJECT_UNCLOAKED,
                self._winevent_callback,
            ),
            hook.hook_winevent(
                WinEvent.EVENT_SYSTEM_MINIMIZESTART,
                WinEvent.EVENT_SYSTEM_MINIMIZEEND,
                self._winevent_callback,
            ),
            hook.hook_winevent(
                WinEvent.EVENT_SYSTEM_MOVESIZEEND,
                WinEvent.EVENT_SYSTEM_MOVESIZEEND,
                partial(self._winevent_callback, restrict=True),
            ),
            # hook.hook_winevent(
            #     WinEvent.EVENT_SYSTEM_FOREGROUND,
            #     WinEvent.EVENT_SYSTEM_FOREGROUND,
            #     partial(self._winevent_callback, restrict=True),
            # ),
        ]

    def uninstall_hooks(self):
        for hook_id in self.hook_ids:
            hook.unhook_winevent(hook_id)
        self.hook_ids = []


if __name__ == "__main__":
    wm = WindowManager()
    wm.sync()
    print()
    wm.sync()



"""

The ``layout`` module operates in Relative Coordinate, to that end, it defines 2 basic types:

``FloatRect`` is a tuple with 4 elements (left/top/right/bottom) to describe a rectangle in ratio form (0.0~1.0)

``Layout`` is a generator which generates FloatRects for given total number of windows

"""

from functools import partial
from typing import Callable, Iterator, Tuple, Union

# FloatRect holds relative coordinate for rectangle (left/top/right/bottom)
FloatRect = Tuple[float, float, float, float]

# Layout accepts an integer (total number of windows) and return a FloatRect generator
Layout = Callable[[int], Iterator[FloatRect]]


def mono(n: int) -> Iterator[FloatRect]:
    """The mono Layout

    .. code-block:: text

        +-----------+
        |           |
        |           |
        |           |
        |     1     |
        |           |
        |           |
        |           |
        +-----------+

    :param n: total number of windows
    :rtype: Iterator[FloatRect]
    """
    for i in range(n):
        yield 0.0, 0.0, 1.0, 1.0


def dwindle(n: int) -> Iterator[FloatRect]:
    """The dwindle Layout

    .. code-block:: text

        +-----------+-----------+
        |           |           |
        |           |     2     |
        |           |           |
        |     1     +-----+-----+
        |           |     |  4  |
        |           |  3  +--+--+
        |           |     | 5|-.|
        +-----------+-----+-----+

    :param n: total number of windows
    :rtype: Iterator[FloatRect]
    """
    l, t, r, b = 0.0, 0.0, 1.0, 1.0
    last_index = n - 1
    for i in range(n):
        # last window would occupy the whole area
        if i == last_index:
            yield l, t, r, b
        # or it should leave out half space for the other windows
        elif i % 2 == 0:
            nl = r - (r - l) / 2
            yield l, t, nl, b
            l = nl
        else:
            nb = b - (b - t) / 2
            yield l, t, r, nb
            t = nb


def static_bigscreen_8(n: int) -> Iterator[FloatRect]:
    """layout for a big screen (like a television) of 55 inches or more. here,
    the 'eye line' should define the upper (main) horizontal segregation. due
    to an attempt keep the eyes below it for main actions on the screen. the
    screen will be optimal for 8 application windows. fewer windows might lead
    to a different layout.

    .. code-block:: text

    +----------+----------+----------+----------+
    |          |          |          |          |
    |          |          |          |          |
    |     6    |     7    |     8    |          |
    |          |          |          |          |
    |          |          |          |          |
    +----------+--+-----+-+----------+          |
    |             |     |            |          |
    |             |     |            |          |
    |             |     |            |          |
    |             |     |            |     4    |
    |             |  3  |      1     |          |
    |     5       |     |            |          |
    |             |     |            |          |
    |             |     |            |          |
    |             +-----+------------|          |
    |             |                  |          |
    |             |         2        |          |
    +-------------+------------------+----------+

    :param n: total number of currently active windows
    :rtype: Iterator[FloatRect]
    """

    # one single window fills the whole screen
    l, t, r, b = 0.0, 0.0, 1.0, 1.0
    h1 = 0.37
    h2 = 0.80
    v1 = 0.30
    v2 = 0.45

    # one window present
    if n==1:
        yield 0.25, 0.37, 0.75, 1.00
    if n==2:
        yield 0.25, 0.37, 0.75, 0.80
        yield 0.25, 0.80, 0.75, 1.00
    if n==3:
        yield 0.45, 0.37, 0.75, 0.80
        yield 0.30, 0.80, 0.75, 1.00
        yield 0.30, 0.37, 0.45, 0.80
    if n==4:
        yield 0.45, 0.37, 0.75, 0.80
        yield 0.30, 0.80, 0.75, 1.00
        yield 0.30, 0.37, 0.45, 0.80
        yield 0.75, 0.00, 1.00, 1.00
    if n==5:
        yield 0.45, 0.37, 0.75, 0.80
        yield 0.30, 0.80, 0.75, 1.00
        yield 0.30, 0.37, 0.45, 0.80
        yield 0.75, 0.00, 1.00, 1.00
        yield 0.00, 0.37, 0.30, 1.00
    if n==6:
        yield 0.45, 0.37, 0.75, 0.80
        yield 0.30, 0.80, 0.75, 1.00
        yield 0.30, 0.37, 0.45, 0.80
        yield 0.75, 0.00, 1.00, 1.00
        yield 0.00, 0.37, 0.30, 1.00
        yield 0.00, 0.00, 0.25, 0.37
    if n==7:
        yield 0.45, 0.37, 0.75, 0.80
        yield 0.30, 0.80, 0.75, 1.00
        yield 0.30, 0.37, 0.45, 0.80
        yield 0.75, 0.00, 1.00, 1.00
        yield 0.00, 0.37, 0.30, 1.00
        yield 0.00, 0.00, 0.25, 0.37
        yield 0.25, 0.00, 0.50, 0.37
    if n==8:
        yield 0.45, 0.37, 0.75, 0.80
        yield 0.30, 0.80, 0.75, 1.00
        yield 0.30, 0.37, 0.45, 0.80
        yield 0.75, 0.00, 1.00, 1.00
        yield 0.00, 0.37, 0.30, 1.00
        yield 0.00, 0.00, 0.25, 0.37
        yield 0.25, 0.00, 0.50, 0.37
        yield 0.50, 0.00, 0.75, 0.37


def widescreen_dwindle(n: int, master_ratio: float = 0.4) -> Iterator[FloatRect]:
    """A wide-screen friendly dwindle Layout

    .. code-block:: text

        +-----------+-----------+-----------+
        |           |           |           |
        |           |           |     3     |
        |           |           |           |
        |     1     |     2     +-----+-----+
        |           |           |     |  5  |
        |           |           |  4  +--+--+
        |           |           |     | 6|-.|
        +-----------+-----------+-----+-----+

    :param n: total number of windows
    :rtype: Iterator[FloatRect]
    """
    if n == 0:
        return
    #    wide_dwindle
    if n == 1:
        yield 0.0, 0.0, 1.0, 1.0
        return
    # master window on the left
    yield 0.0, 0.0, master_ratio, 1
    # other windows on the right with dwindle layout, just map the coordinate and we are good
    yield from map(
        partial(plug_rect, target=(master_ratio, 0.0, 1.0, 1.0)), dwindle(n - 1)
    )


Number = Union[int, float]
NumberRect = Tuple[Number, Number, Number, Number]


def plug_rect(source: FloatRect, target: NumberRect) -> NumberRect:
    """Plug the source rect into the target rect and compute the new dimensions,
    you may plug a Relative Rect into a Physical Rect, but not the other way around.

    :param source: the FloatRect to be moved
    :param target: the container, either a FloatRect or Rect(physical pixels)
    :returns: Rect or FloatRect depends on the type of the target
    :rtype: NumberRect
    """

    sl, st, sr, sb = source
    tl, tt, tr, tb = target
    tw = tr - tl
    th = tb - tt
    return (
        tl + sl * tw,  # left = target left + scaled source left
        tt + st * th,  # top = target top + scaled source top
        tl + sr * tw,  # right
        tt + sb * th,  # bottom
    )


if __name__ == "__main__":
    print("dwindle")
    for n in range(1, 5):
        print(list(dwindle(n)))
    print()
    print("widescreen_dwindle")
    for n in range(1, 5):
        print(list(widescreen_dwindle(n)))



import os
import typing

from PySide6.QtCore import QPoint, Qt
from PySide6.QtGui import QColor, QCursor, QImage, QKeyEvent, QPixmap, QScreen
from PySide6.QtWidgets import (
    QApplication,
    QDialog,
    QFrame,
    QGraphicsDropShadowEffect,
    QHBoxLayout,
    QLabel,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)


class Dialog(QDialog):
    radius: int = 10

    def __init__(
        self,
        radius: int = 15,
        theme: str = "dark",
        spacer: bool = True,
    ):
        super().__init__()
        self.radius = radius
        self.theme = theme

        # window style
        self.setWindowFlag(Qt.WindowType.FramelessWindowHint)
        self.setWindowFlag(Qt.WindowType.WindowStaysOnTopHint)
        self.setWindowFlag(Qt.WindowType.Tool)
        self.setWindowFlag(Qt.WindowType.Popup)
        self.setAttribute(Qt.WidgetAttribute.WA_NoSystemBackground, True)
        self.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose, True)
        self.setModal(True)

        # create common widgets
        # self.central = QWidget(self)
        # self.central.setObjectName("central")
        # self.central_layout = QVBoxLayout(self.central)
        # self.setCentralWidget(self.central)
        self.central = self
        self.central_layout = QVBoxLayout(self)

        self.root = QWidget(self.central)
        self.root.setObjectName("root")
        self.root.setSizePolicy(
            QSizePolicy.Policy.Expanding,
            QSizePolicy.Policy.Minimum if spacer else QSizePolicy.Policy.Expanding,
        )
        self.root_layout = QVBoxLayout(self.root)
        self.container = QFrame(self.root)
        self.container.setObjectName("container")
        self.container.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum
        )
        self.container_layout = QVBoxLayout(self.container)
        self.root_layout.addWidget(self.container)
        self.central_layout.addWidget(self.root)
        if spacer:
            self.spacer = QWidget(self)
            self.spacer.setSizePolicy(
                QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding
            )
            self.central_layout.addWidget(self.spacer)

        # setup common widgets
        self.central_layout.setContentsMargins(
            self.radius, self.radius, self.radius, self.radius
        )
        self.root_layout.setContentsMargins(0, self.radius, 0, self.radius)
        self.container_layout.setContentsMargins(0, 0, 0, 0)
        self.container_layout.setSpacing(0)

        # drop shadow
        self.root_shadow = QGraphicsDropShadowEffect()
        self.root_shadow.setBlurRadius(self.radius)
        self.root_shadow.setColor(QColor(0, 0, 0, 100))
        self.root.setGraphicsEffect(self.root_shadow)

        # load theme
        with open(
            os.path.join(
                os.path.dirname(os.path.dirname(__file__)),
                "assets",
                f"{self.theme}.css",
            ),
            "r",
            encoding="utf-8",
        ) as f:
            self.setStyleSheet(f.read())

    def screenFromCursor(self) -> typing.Tuple[QScreen, QPoint]:
        pos = QCursor.pos()
        app = QApplication.instance()
        for screen in app.screens():
            if screen.geometry().contains(pos):
                return screen

    def keyPressEvent(self, a0: QKeyEvent) -> None:
        if a0.key() == Qt.Key.Key_Escape:
            self.onEscPresssed()
            return
        return super().keyPressEvent(a0)

    def onEscPresssed(self) -> None:
        self.hide()

    @staticmethod
    def deleteDirectChildren(
        container: QWidget,
        keep: typing.Optional[typing.Callable[[QWidget], bool]] = None,
    ):
        for w in container.findChildren(
            QWidget, options=Qt.FindChildOption.FindDirectChildrenOnly
        ):
            if keep and keep(w):
                continue
            w.deleteLater()

    def create_row_widget(self, width: int, text: str, icon: QImage = None) -> QWidget:
        """create a QtWidget for displaying Shortcut"""
        frame_widget = QFrame()
        frame_widget.setObjectName("menuitem")
        frame_layout = QHBoxLayout(frame_widget)
        frame_layout.setContentsMargins(5, 2, 5, 2)
        frame_layout.setSpacing(10)
        width -= 10
        # icon
        icon_widget = QLabel(frame_widget)
        icon_widget.setObjectName("icon")
        icon_widget.setFixedSize(32, 32)
        width -= 32
        icon_widget.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        frame_layout.addWidget(icon_widget, stretch=0)
        if icon:
            icon_widget.setPixmap(QPixmap.fromImage(icon))
            icon_widget.setScaledContents(True)
            icon_widget.setAlignment(Qt.AlignmentFlag.AlignTop)
        # text
        content_widget = QLabel(frame_widget)
        content_widget.setObjectName("content")
        content_widget.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding
        )
        content_widget.setText(text)
        frame_layout.addWidget(content_widget)
        return frame_widget
